/****************************************************************************
** libmatroska : parse Matroska files, see http://www.matroska.org/
**
** <file/class description>
**
** Copyright (C) 2002-2003 Christophe Paris.  All rights reserved.
**
** This file may be distributed under the terms of the Q Public License
** as defined by Trolltech AS of Norway and appearing in the file
** LICENSE.QPL included in the packaging of this file.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding an other license may use this file in accordance with 
** the Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.matroska.org/license/qpl/ for QPL licensing information.
** See http://www.matroska.org/license/gpl/ for GPL licensing information.
**
** Contact license@matroska.org if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
/*!
	\file
	\version \$Id: MatroskaReader.cpp,v 1.17 2003/05/11 12:13:35 robux4 Exp $
	\author Christophe Paris     <toffparis @ users.sf.net>
*/

#include "MatroskaReader.h"
#include <dshow.h>
#include <wxdebug.h>
#include "EbmlVoid.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MatroskaReader::MatroskaReader()
 :bReadDummyElements(true)
 ,m_pStream(NULL)
 ,m_pKaxIO(NULL)
 ,m_TrackCount(0)
 ,m_qwDuration(0)
 ,m_TimeScale(10000.0)
 ,m_CueEntries(NULL)
 ,m_MetaSeeks(NULL)
{
}

// ----------------------------------------------------------------------------

MatroskaReader::~MatroskaReader()
{
	KaxTrackInfoStruct* pTrackInfo;
	for(int i=0; i<m_TrackCount; i++)
	{
		pTrackInfo = m_Tracks[i];
		if (pTrackInfo->CodecID != NULL) 
			free(pTrackInfo->CodecID);
		if (pTrackInfo->CodecPrivate != NULL) 
			free(pTrackInfo->CodecPrivate);
		if (pTrackInfo->Type == track_video)
			delete pTrackInfo->pVideoInfo;
		if (pTrackInfo->Type == track_audio) 
			delete pTrackInfo->pAudioInfo;
	}
	
	if (m_pKaxIO != NULL)
		m_pKaxIO->close();

	if (m_CueEntries != NULL)
		delete m_CueEntries;

	if (m_MetaSeeks != NULL)
		delete m_MetaSeeks;
}

// ----------------------------------------------------------------------------

int MatroskaReader::InitKaxFile(IOCallback *pIOCb) 
{
	bool	bTracksLoaded = false;
	int		UpperElementLevel = 0;
	
    NOTE("MatroskaReader::InitKaxFile");
	
	// initialize Element array
	for (int i=0; i<6; i++)
		m_pElems[i] = NULL;
	
	// initialize EBML Stream
	if (pIOCb == NULL)
		return -1;
	
	if (m_pStream != NULL)
		return -2;	// for now we dont support this
	
	m_CueEntries = NULL;
	m_MetaSeeks = NULL;

	if ((m_pStream = new EbmlStream(*pIOCb)) == NULL)
		return -3;
	
	m_pKaxIO = pIOCb;
	
	// is this a valid EBML file ?
	if ((m_pElems[0] = m_pStream->FindNextID(EbmlHead::ClassInfos, 0xFFFFFFFFL)) != NULL)
	{
		// Make sure we have the expected element		
		if (!(EbmlId(*m_pElems[0]) == EbmlHead::ClassInfos.GlobalId))
		{
			return -1;
		}
		
		m_pElems[1] = m_pStream->FindNextElement(m_pElems[0]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
		while (m_pElems[1] != NULL) 
		{
			if (UpperElementLevel != 0)
				break;
			
			if (EbmlId(*m_pElems[1]) == EDocType::ClassInfos.GlobalId)
			{
				EDocType & DocType = *static_cast<EDocType*>(m_pElems[1]);
				DocType.ReadData(m_pStream->I_O());
				if (std::string(DocType) != "matroska")
					return -5;
				break; // we are finished with the EbmlHead for now
			}

			if (UpperElementLevel > 0) 
			{
				UpperElementLevel--;
				delete m_pElems[1];
				m_pElems[1] = m_pElems[2];
				if (UpperElementLevel > 0)
					break;
			}
			else
			{
				m_pElems[1]->SkipData(*m_pStream, m_pElems[1]->Generic().Context);
				delete m_pElems[1];
				
				m_pElems[1] = m_pStream->FindNextElement(m_pElems[0]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
			}
		}
		
		m_pElems[0]->SkipData(*m_pStream, EbmlHead_Context);
		if (m_pElems[0] != NULL)
		{
			delete m_pElems[0];
			m_pElems[0] = NULL;
		}
	}
	else
	{
		return -4;
	}
	
	// find a matroska segment
	if ((m_pElems[0] = m_pStream->FindNextID(KaxSegment::ClassInfos, 0xFFFFFFFFL)) != NULL)
		if (!(EbmlId(*m_pElems[0]) == KaxSegment::ClassInfos.GlobalId))
			return -5;
		
		// parse all sub-elements
		m_pElems[1] = m_pStream->FindNextElement(m_pElems[0]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
		while (m_pElems[1] != NULL) 
		{
			if (UpperElementLevel != 0)
				break;
			
			if (EbmlId(*m_pElems[1]) == KaxTracks::ClassInfos.GlobalId)
			{
				// another funky level ;-)
				m_pElems[2] = m_pStream->FindNextElement(m_pElems[1]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
				while (m_pElems[2] != NULL) 
				{
					if (UpperElementLevel != 0)
						break;
					if (EbmlId(*m_pElems[2]) == KaxTrackEntry::ClassInfos.GlobalId)
					{					
						// and fill it with those values					
						KaxTrackInfoStruct *TrackInfo = new KaxTrackInfoStruct;
						memset(TrackInfo, 0, sizeof(KaxTrackInfoStruct));						
						InitTrack(m_pElems, UpperElementLevel, TrackInfo);
						m_Tracks[m_TrackCount] = TrackInfo;
						m_TrackCount++;
					}
					
					if (UpperElementLevel > 0) 
					{
						UpperElementLevel--;
						delete m_pElems[2];
						m_pElems[2] = m_pElems[3];
						if (UpperElementLevel > 0)
							break;
					}
					else
					{
						m_pElems[2]->SkipData(*m_pStream, m_pElems[2]->Generic().Context);
						delete m_pElems[2];
						
						m_pElems[2] = m_pStream->FindNextElement(m_pElems[1]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
					}
				}
				
				bTracksLoaded = true;
			}
			else if (EbmlId(*m_pElems[1]) == KaxCluster::ClassInfos.GlobalId) 
			{
				// store position
				m_qwKaxSegStart = m_pKaxIO->getFilePointer();
				
				m_pTmpElems[0] = m_pElems[0];
				m_pTmpElems[1] = m_pElems[1];
				
				// and exit if we already parsed the TrackEntry
				m_iLevel = 1;
				m_iTmpUpElLev = m_iUpElLev = UpperElementLevel;
				
				if (bTracksLoaded)
					return 1;
			}
			else if (EbmlId(*m_pElems[1]) == KaxInfo::ClassInfos.GlobalId) 
			{
				m_pElems[2] = m_pStream->FindNextElement(m_pElems[1]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
				while (m_pElems[2] != NULL)
				{
					if (UpperElementLevel != 0)
						break;
					if (EbmlId(*m_pElems[2]) == KaxTimecodeScale::ClassInfos.GlobalId) 
					{
						KaxTimecodeScale & TimecodeScale = *static_cast<KaxTimecodeScale*>(m_pElems[2]);
						TimecodeScale.ReadData(*m_pKaxIO);
						uint32 timescale = TimecodeScale;
						m_TimeScale = (double)timescale;
						m_TimeScale /= 100.0f;	// multiplicator (convert 1ns -> 100ns)
					}
					else if (EbmlId(*m_pElems[2]) == KaxDuration::ClassInfos.GlobalId) 
					{
						KaxDuration & Duration = *static_cast<KaxDuration*>(m_pElems[2]);
						Duration.ReadData(*m_pKaxIO);
						m_qwDuration = ((float)Duration) * m_TimeScale;
					}
					
					if (UpperElementLevel > 0) 
					{
						UpperElementLevel--;
						delete m_pElems[2];
						m_pElems[2] = m_pElems[3];
						if (UpperElementLevel > 0)
							break;
					}
					else 
					{
						m_pElems[2]->SkipData(*m_pStream, m_pElems[2]->Generic().Context);
						delete m_pElems[2];
						
						m_pElems[2] = m_pStream->FindNextElement(m_pElems[1]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
					}
				}
			}
			
			if (UpperElementLevel > 0) 
			{
				UpperElementLevel--;
				delete m_pElems[1];
				m_pElems[1] = m_pElems[2];
				if (UpperElementLevel > 0)
					break;
			}
			else
			{
				m_pElems[1]->SkipData(*m_pStream, m_pElems[1]->Generic().Context);
				delete m_pElems[1];
				
				m_pElems[1] = m_pStream->FindNextElement(m_pElems[0]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
			}
	}
	
	return 0;
}

// ----------------------------------------------------------------------------

// parsing KaxTrackEntry SubElements
int MatroskaReader::InitTrack(EbmlElement **elems, int &UEL, KaxTrackInfoStruct *pTrackInfos)
{
	EbmlElement			*plev3, *plev4, *plev5;
	int					UpElLev = UEL;
	
    NOTE("MatroskaReader::InitTrack");
	
	plev3 = plev4 = plev5 = NULL;
	
	// PrE - computer video player profile
	while (1)
	{
		if ((plev3 = m_pStream->FindNextElement(elems[2]->Generic().Context, UpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL) 
			break;
		
		if (UpElLev != 0) 
			break;
		
		if (EbmlId(*plev3) == KaxTrackNumber::ClassInfos.GlobalId)
		{
			KaxTrackNumber &TrackNum = *static_cast<KaxTrackNumber*>(plev3);
			TrackNum.ReadData(*m_pKaxIO);
			pTrackInfos->Number = TrackNum;
		}
		else if (EbmlId(*plev3) == KaxTrackType::ClassInfos.GlobalId)
		{
			KaxTrackType &TrackType = *static_cast<KaxTrackType*>(plev3);
			TrackType.ReadData(*m_pKaxIO);
			pTrackInfos->Type = TrackType;
		}
		else if (EbmlId(*plev3) == KaxTrackFlagEnabled::ClassInfos.GlobalId)
		{
			KaxTrackFlagEnabled &FlagEnabled = *static_cast<KaxTrackFlagEnabled*>(plev3);
			FlagEnabled.ReadData(*m_pKaxIO);
			pTrackInfos->Enabled = FlagEnabled;
		}
		else if (EbmlId(*plev3) == KaxTrackFlagDefault::ClassInfos.GlobalId) 
		{
			KaxTrackFlagDefault &FlagDefault = *static_cast<KaxTrackFlagDefault*>(plev3);
			FlagDefault.ReadData(*m_pKaxIO);
			pTrackInfos->Default = FlagDefault;
		}
		else if (EbmlId(*plev3) == KaxTrackFlagLacing::ClassInfos.GlobalId)
		{
			KaxTrackFlagLacing &FlagLacing= *static_cast<KaxTrackFlagLacing*>(plev3);
			FlagLacing.ReadData(*m_pKaxIO);
			pTrackInfos->Lacing = FlagLacing;
		}
		else if (EbmlId(*plev3) == KaxTrackMinCache::ClassInfos.GlobalId) 
		{
			KaxTrackMinCache &MinCache = *static_cast<KaxTrackMinCache*>(plev3);
			MinCache.ReadData(*m_pKaxIO);
			pTrackInfos->MinCache = MinCache;
		}
		else if (EbmlId(*plev3) == KaxTrackMaxCache::ClassInfos.GlobalId) 
		{
			KaxTrackMaxCache &MaxCache = *static_cast<KaxTrackMaxCache*>(plev3);
			MaxCache.ReadData(*m_pKaxIO);
			pTrackInfos->MaxCache = MaxCache;
		}
		else if (EbmlId(*plev3) == KaxTrackName::ClassInfos.GlobalId) 
		{
			KaxTrackName &Name = *static_cast<KaxTrackName*>(plev3);
			Name.ReadData(*m_pKaxIO);
			pTrackInfos->Name = Name;
		}
		else if (EbmlId(*plev3) == KaxTrackLanguage::ClassInfos.GlobalId) 
		{
			KaxTrackLanguage &Language = *static_cast<KaxTrackLanguage*>(plev3);
			Language.ReadData(*m_pKaxIO);
			pTrackInfos->Language = Language;
		}
		else if (EbmlId(*plev3) == KaxCodecID::ClassInfos.GlobalId) 
		{
			KaxCodecID &CodecID = *static_cast<KaxCodecID*>(plev3);
			CodecID.ReadData(*m_pKaxIO);
			pTrackInfos->CodecID = (binary*)malloc(CodecID.GetSize() * sizeof(binary));
			memcpy(pTrackInfos->CodecID, CodecID.GetBuffer(), CodecID.GetSize());
		}
		else if (EbmlId(*plev3) == KaxCodecPrivate::ClassInfos.GlobalId)
		{
			KaxCodecPrivate &CodecPrivate = *static_cast<KaxCodecPrivate*>(plev3);
			CodecPrivate.ReadData(*m_pKaxIO);
			pTrackInfos->CodecPrivateLen = CodecPrivate.GetSize();
			pTrackInfos->CodecPrivate = (binary*)malloc(CodecPrivate.GetSize() * sizeof(binary));
			memcpy(pTrackInfos->CodecPrivate, CodecPrivate.GetBuffer(), CodecPrivate.GetSize());			
		}
		else if (EbmlId(*plev3) == KaxCodecName::ClassInfos.GlobalId) 
		{
			KaxCodecName &CodecName = *static_cast<KaxCodecName*>(plev3);
			CodecName.ReadData(*m_pKaxIO);
			pTrackInfos->CodecName = CodecName;
		}
		else if (EbmlId(*plev3) == KaxCodecDecodeAll::ClassInfos.GlobalId)
		{
			KaxCodecDecodeAll &CodecDecodeAll = *static_cast<KaxCodecDecodeAll*>(plev3);
			CodecDecodeAll.ReadData(*m_pKaxIO);
			pTrackInfos->CodecDecodeAll = CodecDecodeAll;
		}
		else if (EbmlId(*plev3) == KaxTrackOverlay::ClassInfos.GlobalId) 
		{
			KaxTrackOverlay &Overlay = *static_cast<KaxTrackOverlay*>(plev3);
			Overlay.ReadData(*m_pKaxIO);
			pTrackInfos->Overlay = Overlay;
		}
		else if (EbmlId(*plev3) == KaxTrackVideo::ClassInfos.GlobalId) 
		{
			pTrackInfos->pVideoInfo = new KaxVideoTrackInfoStruct;
			memset(pTrackInfos->pVideoInfo, 0, sizeof(KaxVideoTrackInfoStruct));
			
			while (1) 
			{
				if ((plev4 = m_pStream->FindNextElement(plev3->Generic().Context, UpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL) 
					break;
				
				if (UpElLev != 0)
					break;
				
				if (EbmlId(*plev4) == KaxVideoFlagInterlaced::ClassInfos.GlobalId) 
				{
					KaxVideoFlagInterlaced &Interlaced = *static_cast<KaxVideoFlagInterlaced*>(plev4);
					Interlaced.ReadData(*m_pKaxIO);
					pTrackInfos->pVideoInfo->Interlaced = Interlaced;
				}
				else if (EbmlId(*plev4) == KaxVideoStereoMode::ClassInfos.GlobalId)
				{
					KaxVideoStereoMode &StereoMode = *static_cast<KaxVideoStereoMode*>(plev4);
					StereoMode.ReadData(*m_pKaxIO);
					pTrackInfos->pVideoInfo->StereoMode = StereoMode;
				}
				else if (EbmlId(*plev4) == KaxVideoPixelWidth::ClassInfos.GlobalId)
				{
					KaxVideoPixelWidth &PixelWidth = *static_cast<KaxVideoPixelWidth*>(plev4);
					PixelWidth.ReadData(*m_pKaxIO);
					pTrackInfos->pVideoInfo->PixelWidth = PixelWidth;
				}
				else if (EbmlId(*plev4) == KaxVideoPixelHeight::ClassInfos.GlobalId)
				{
					KaxVideoPixelHeight &PixelHeight = *static_cast<KaxVideoPixelHeight*>(plev4);
					PixelHeight.ReadData(*m_pKaxIO);
					pTrackInfos->pVideoInfo->PixelHeight = PixelHeight;
				}
				
				if (UpElLev > 0) 
				{
					UpElLev--;
					delete plev4;
					plev4 = plev5;
					if (UpElLev > 0) 
						break;
				}
				else {
					plev4->SkipData(*m_pStream, plev4->Generic().Context);
					delete plev4;
				}
			}
		}
		else if (EbmlId(*plev3) == KaxTrackAudio::ClassInfos.GlobalId)
		{
			pTrackInfos->pAudioInfo = new KaxAudioTrackInfoStruct;
			memset(pTrackInfos->pAudioInfo, 0, sizeof(KaxAudioTrackInfoStruct));
			
			while (1)
			{
				if ((plev4 = m_pStream->FindNextElement(plev3->Generic().Context, UpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL) break;
				if (UpElLev != 0) break;
				
				if (EbmlId(*plev4) == KaxAudioSamplingFreq::ClassInfos.GlobalId) 
				{
					KaxAudioSamplingFreq &SampFreq = *static_cast<KaxAudioSamplingFreq*>(plev4);
					SampFreq.ReadData(*m_pKaxIO);
					pTrackInfos->pAudioInfo->SamplingFreq = SampFreq;
				}
				else if (EbmlId(*plev4) == KaxAudioChannels::ClassInfos.GlobalId)
				{
					KaxAudioChannels &Channels = *static_cast<KaxAudioChannels*>(plev4);
					Channels.ReadData(*m_pKaxIO);
					pTrackInfos->pAudioInfo->Channels = Channels;
				}
				else if (EbmlId(*plev4) == KaxAudioPosition::ClassInfos.GlobalId)
				{
					KaxAudioPosition &ChanPos = *static_cast<KaxAudioPosition*>(plev4);
					ChanPos.ReadData(*m_pKaxIO);
					pTrackInfos->pAudioInfo->Position = ChanPos.GetBuffer();
				}
				else if (EbmlId(*plev4) == KaxAudioBitDepth::ClassInfos.GlobalId) 
				{
					KaxAudioBitDepth &BitDepth = *static_cast<KaxAudioBitDepth*>(plev4);
					BitDepth.ReadData(*m_pKaxIO);
					pTrackInfos->pAudioInfo->BitDepth = BitDepth;
					}/*
					 else if (EbmlId(*plev4) == KaxAudioSubTrack::ClassInfos.GlobalId) {
					 
					}*/
				
				if (UpElLev > 0) 
				{
					UpElLev--;
					delete plev4;
					plev4 = plev5;
					if (UpElLev > 0)
						break;
				}
				else {
					plev4->SkipData(*m_pStream, plev4->Generic().Context);
					delete plev4;
				}
			}
		}
		
		if (UpElLev > 0) 
		{
			UpElLev--;
			delete plev3;
			plev3 = plev4;
			if (UpElLev > 0)
				break;
		}
		else
		{
			plev3->SkipData(*m_pStream, plev3->Generic().Context);
			delete plev3;
		}
	}
	
	elems[3] = plev3;
	elems[4] = plev4;
	elems[5] = plev5;
	
	UEL = UpElLev;
	return 0;
}

// ----------------------------------------------------------------------------

int MatroskaReader::GetTrackNumber() const
{
	return m_TrackCount;
}

// ----------------------------------------------------------------------------

KaxTrackInfoStruct* MatroskaReader::GetTrackInfo(int i)
{
	return m_Tracks[i];
}

// ----------------------------------------------------------------------------

KaxBlock *MatroskaReader::GetNextBlock() 
{
	KaxCluster	*SegCluster;
	KaxBlock	*pBlock;
	
	switch (m_iLevel) 
	{
	case -1: goto bgroup;
	case 0: break;
	case 1: goto level1;
	case 2: goto level2;
	case 3: goto level3;
	default: return NULL;
	}
	
	m_iLevel = 0;
	if ((m_pElems[0] = m_pStream->FindNextID(KaxSegment::ClassInfos, 0xFFFFFFFFL)) == NULL)
		return NULL;
	
	while (1) 
	{
		if (m_iUpElLev != 0) break;
		
		// segment (level 0)
		if (EbmlId(*m_pElems[0]) == KaxSegment::ClassInfos.GlobalId)
		{
			m_iLevel = 1;
			
			if ((m_pElems[1] = m_pStream->FindNextElement(m_pElems[0]->Generic().Context, m_iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL)
				break;

			while (1)
			{
				if (m_iUpElLev != 0)
					break;
level1:
				// cluster (level 1)
				if (EbmlId(*m_pElems[1]) == KaxCluster::ClassInfos.GlobalId) 
				{
					SegCluster = static_cast<KaxCluster *>(m_pElems[1]);
					
					m_iLevel = 2;
					if ((m_pElems[2] = m_pStream->FindNextElement(m_pElems[1]->Generic().Context, m_iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL)
						break;

					while (1)
					{
						if (m_iUpElLev != 0) break;
level2:
						if (EbmlId(*m_pElems[2]) == KaxClusterTimecode::ClassInfos.GlobalId)
						{
							KaxClusterTimecode &ClusterTime = *static_cast<KaxClusterTimecode*>(m_pElems[2]);
							ClusterTime.ReadData(*m_pKaxIO);
							SegCluster->InitTimecode(uint32(ClusterTime));
						}
						else if (EbmlId(*m_pElems[2]) == KaxBlockGroup::ClassInfos.GlobalId)
						{							
							m_iLevel = 3;
							if ((m_pElems[3] = m_pStream->FindNextElement(m_pElems[2]->Generic().Context, m_iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL) 
								break;

							while (1) 
							{
								if (m_iUpElLev != 0) 
									break;
level3:
								// for now we're only interested in BLOX
								if (EbmlId(*m_pElems[3]) == KaxBlock::ClassInfos.GlobalId) 
								{
									pBlock = static_cast<KaxBlock*>(m_pElems[3]);
									pBlock->ReadData(*m_pKaxIO);
									pBlock->SetParent(*((KaxCluster*)m_pElems[1]));
									m_iLevel = -1;	// means we're in blockgroup
									return pBlock;
								}
								else if (EbmlId(*m_pElems[3]) == KaxBlockDuration::ClassInfos.GlobalId)
								{
									// TODO : we need the BlockDuration for subtitles
									KaxBlockDuration &kaxBlockDuration = *static_cast<KaxBlockDuration*>(m_pElems[3]);
									/*
									kaxBlockDuration.ReadData(*m_pKaxIO);
									uint32 kbdur =  uint32(kaxBlockDuration) * TimeScale;
									*/
								}
bgroup:
								if (m_iUpElLev > 0) 
								{
									m_iUpElLev--;
									
									if (!(EbmlId(*m_pElems[3]) == KaxBlock::ClassInfos.GlobalId))
										delete m_pElems[3];		// done inside outpin's deliverblock()
									
									m_pElems[3] = m_pElems[4];
									if (m_iUpElLev > 0) 
										break;
								}
								else 
								{
									m_pElems[3]->SkipData(*m_pStream, m_pElems[3]->Generic().Context);
									
									if (!(EbmlId(*m_pElems[3]) == KaxBlock::ClassInfos.GlobalId))
										delete m_pElems[3];		// done inside outpin's deliverblock()
									
									if ((m_pElems[3] = m_pStream->FindNextElement(m_pElems[2]->Generic().Context, m_iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL)
										return NULL;		// xyx - not that good - NULL may be signal for unknown element
								}
							}
						}
						
						if (m_iUpElLev > 0)
						{
							m_iUpElLev--;
							delete m_pElems[2];
							m_pElems[2] = m_pElems[3];
							if (m_iUpElLev > 0)
								break;
						}
						else 
						{
							m_pElems[2]->SkipData(*m_pStream, m_pElems[2]->Generic().Context);
							delete m_pElems[2];
							
							if ((m_pElems[2] = m_pStream->FindNextElement(m_pElems[1]->Generic().Context, m_iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL)
								break;
						}
					}
				}
				/*
				else if (m_pElems[1]->IsDummy())//EbmlId(*m_pElems[1]) == DummyRawElement::ClassInfos.GlobalId)
				break;//continue;	// skip() = assert() !!!
				*/		
				if (m_iUpElLev > 0) 
				{
					m_iUpElLev--;
					delete m_pElems[1];
					m_pElems[1] = m_pElems[2];
					if (m_iUpElLev > 0)
						break;
				}
				else 
				{
					m_pElems[1]->SkipData(*m_pStream, m_pElems[1]->Generic().Context);
					delete m_pElems[1];
					
					if ((m_pElems[1] = m_pStream->FindNextElement(m_pElems[0]->Generic().Context, m_iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL)
						break;
				}
			}
		}
		
		if (m_iUpElLev > 0)
		{
			m_iUpElLev--;
			delete m_pElems[0];
			m_pElems[0] = m_pElems[1];
			if (m_iUpElLev > 0)
				break;
		}
		else {
			m_pElems[0]->SkipData(*m_pStream, m_pElems[0]->Generic().Context);
			delete m_pElems[0];
			
			if ((m_pElems[0] = m_pStream->FindNextID(KaxSegment::ClassInfos, 0xFFFFFFFFL)) == NULL)
				break;
		}
	}
	
	return NULL;
}

// ----------------------------------------------------------------------------

/*!
	\todo put element in a home made Cue entry if none is found in the file
*/
KaxBlockGroup *MatroskaReader::GetNextBlockG() 
{
	KaxCluster	  *SegCluster;
	KaxBlockGroup *pBlockG;
	bool bKeepElement1 = false;
	bool bKeepElement2 = true;

	switch (m_iLevel) 
	{
	case -2: goto bgroup;
	case 0: break;
	case 1: goto level1;
	case 2: goto level2;
	case 10:goto levelSeek;
	default: return NULL;
	}
	
	m_iLevel = 0;
	if ((m_pElems[0] = m_pStream->FindNextID(KaxSegment::ClassInfos, 0xFFFFFFFFL)) == NULL)
		return NULL;
	
	while (1) 
	{
		if (m_iUpElLev != 0) break;
		
		// segment (level 0)
		if (EbmlId(*m_pElems[0]) == KaxSegment::ClassInfos.GlobalId)
		{
			m_iLevel = 1;
levelSeek:			
			if ((m_pElems[1] = m_pStream->FindNextElement(KaxSegment::ClassInfos.Context, m_iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL)
				break;

			while (1)
			{
				if (m_iUpElLev != 0)
					break;
level1:
				// cluster (level 1)
				if (EbmlId(*m_pElems[1]) == KaxCluster::ClassInfos.GlobalId) 
				{
					SegCluster = static_cast<KaxCluster *>(m_pElems[1]);
					
					m_iLevel = 2;
					if ((m_pElems[2] = m_pStream->FindNextElement(m_pElems[1]->Generic().Context, m_iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL)
						break;

					while (1)
					{
						if (m_iUpElLev != 0) break;
level2:
						if (EbmlId(*m_pElems[2]) == KaxClusterTimecode::ClassInfos.GlobalId)
						{
							bKeepElement2 = false;
							KaxClusterTimecode &ClusterTime = *static_cast<KaxClusterTimecode*>(m_pElems[2]);
							ClusterTime.ReadData(*m_pKaxIO);
							SegCluster->InitTimecode(uint32(ClusterTime));
						}
						else if (EbmlId(*m_pElems[2]) == KaxBlockGroup::ClassInfos.GlobalId)
						{							
							pBlockG = static_cast<KaxBlockGroup*>(m_pElems[2]);
							pBlockG->SetParent(*((KaxCluster*)m_pElems[1]));
							pBlockG->Read(*m_pStream, KaxBlockGroup::ClassInfos.Context, m_iUpElLev, m_pElems[3], bReadDummyElements);
							KaxBlock *TheBlock = static_cast<KaxBlock *>(pBlockG->FindElt(KaxBlock::ClassInfos));
							if (TheBlock != NULL) {
								TheBlock->SetParent(*((KaxCluster*)m_pElems[1]));
							}
							m_iLevel = -2;	// means we're in blockgroup
							return pBlockG;
						} else {
							bKeepElement2 = false;
						}
bgroup:						
						if (m_iUpElLev > 0)
						{
							m_iUpElLev--;
							if (!bKeepElement2)
								delete m_pElems[2];
							m_pElems[2] = m_pElems[3];
							if (m_iUpElLev > 0)
								break;
						}
						else 
						{
							m_pElems[2]->SkipData(*m_pStream, m_pElems[2]->Generic().Context);
							if (!bKeepElement2)
								delete m_pElems[2];
							
							if ((m_pElems[2] = m_pStream->FindNextElement(m_pElems[1]->Generic().Context, m_iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL)
								break;
						}
					}
					bKeepElement1 = false;
				} else if (EbmlId(*m_pElems[1]) == KaxCues::ClassInfos.GlobalId) {
					if (m_CueEntries != NULL) {
						///< should we merge the two ?
					} else {
						m_CueEntries = static_cast<KaxCues*>(m_pElems[1]);
						m_CueEntries->Read(*m_pStream, KaxCues::ClassInfos.Context, m_iUpElLev, m_pElems[2], bReadDummyElements);
						m_CueEntries->Sort();
						bKeepElement1 = true;
					}
				} else if (EbmlId(*m_pElems[1]) == KaxSeekHead::ClassInfos.GlobalId) {
					if (m_MetaSeeks != NULL) {
						///< should we merge the two ?
					} else {
						m_MetaSeeks = static_cast<KaxSeekHead*>(m_pElems[1]);
						m_MetaSeeks->Read(*m_pStream, KaxSeekHead::ClassInfos.Context, m_iUpElLev, m_pElems[2], bReadDummyElements);
						m_CueEntries = GetCueEntry(*m_pStream, *m_MetaSeeks, *static_cast<KaxSegment *>(m_pElems[0]));
						if (m_CueEntries != NULL)
							m_CueEntries->Sort();
						bKeepElement1 = true;
					}
				} else {
					bKeepElement1 = false;
				}

				/*
				else if (m_pElems[1]->IsDummy())//EbmlId(*m_pElems[1]) == DummyRawElement::ClassInfos.GlobalId)
				break;//continue;	// skip() = assert() !!!
				*/		
				if (m_iUpElLev > 0) 
				{
					m_iUpElLev--;
					if (!bKeepElement1)
						delete m_pElems[1];
					m_pElems[1] = m_pElems[2];
					if (m_iUpElLev > 0)
						break;
				}
				else 
				{
					///< \todo modify FindNextElement to disable using global elements instead, in seeking mode
					if (!(EbmlId(*m_pElems[1]) == EbmlVoid::ClassInfos.GlobalId)) {
						m_pElems[1]->SkipData(*m_pStream, m_pElems[1]->Generic().Context);
					}
					if (!bKeepElement1)
						delete m_pElems[1];
					
					if ((m_pElems[1] = m_pStream->FindNextElement(m_pElems[0]->Generic().Context, m_iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL)
						break;
				}
			}
		}
		
		if (m_iUpElLev > 0)
		{
			m_iUpElLev--;
			delete m_pElems[0];
			m_pElems[0] = m_pElems[1];
			if (m_iUpElLev > 0)
				break;
		}
		else {
			m_pElems[0]->SkipData(*m_pStream, m_pElems[0]->Generic().Context);
			delete m_pElems[0];
			
			if ((m_pElems[0] = m_pStream->FindNextID(KaxSegment::ClassInfos, 0xFFFFFFFFL)) == NULL) {
				// no more data in this file
				m_iLevel = 0;
				break;
			}
		}
	}
	
	return NULL;
}

// ----------------------------------------------------------------------------

void  MatroskaReader::Reset()
{
	m_pElems[0] = 0;
	m_pElems[1] = 0;
	m_iUpElLev = 0;
	m_iLevel = 0;
	m_pKaxIO->setFilePointer(0);	
}

// ----------------------------------------------------------------------------

void MatroskaReader::JumpToTimecode(uint64 aTimecode)
{
	int approx_location = 0; ///< \todo position of the first Cluster
	KaxSegment * aSegment = static_cast<KaxSegment *>(m_pElems[0]);

	// cue mode
	if (m_CueEntries != NULL)
	{
		approx_location = m_CueEntries->GetTimecodePosition(aTimecode * 100);
	}

	// blind mode
	else
	{
		///< \todo alternative version with the length of the Segment
		int64 SegLength;
		int64 CurLocation = m_pKaxIO->getFilePointer();
		if (aSegment != NULL && aSegment->IsFiniteSize()) {
			SegLength = aSegment->GetSize();
		} else {
			m_pKaxIO->setFilePointer(0, seek_end);
			SegLength = m_pKaxIO->getFilePointer();
		}
		
		if (m_qwDuration != 0) {
			// dirty
			int64 duration = m_qwDuration, timecode = aTimecode;
			approx_location = uint64(double(SegLength) * double(timecode) / double(duration));
		}

		// meta seek mode
		if (m_MetaSeeks != NULL)
		{
			// find the Cluster closer to this location
			approx_location = FindClusterLocated(approx_location, *m_MetaSeeks);
		}
	}

	if (m_iLevel == -1) {
		m_iLevel = 3;
	}
	if (m_iLevel == -2) {
		m_iLevel = 2;
	}
	if (m_iLevel != 10) {
		// reset all read tags
		for (UINT i=1; i<=m_iLevel; i++) {
			if (m_pElems[i] != NULL) {
//				delete m_pElems[i]; // can't delete an element if it's a saved one Block, 
				m_pElems[i] = NULL;
			}
		}
	}
	m_iLevel = 10; // levelSeek
	m_iUpElLev = 0;

	// set the file pointer to the most appropriate location
	if (aSegment != NULL) {
		NOTE1("Seek at location %I64d in the file", aSegment->GetGlobalPosition(approx_location));
		m_pKaxIO->setFilePointer(aSegment->GetGlobalPosition(approx_location));
	} else {
		NOTE1("Seek to timecode %I64d but no Segment Found !!! (we are probably at the end of the file)", aTimecode);
		m_pKaxIO->setFilePointer(approx_location);
	}
}

/*!
	\brief find a Cue entry in the meta seek and read it in memory
*/
KaxCues * MatroskaReader::GetCueEntry(EbmlStream & inDataStream, const KaxSeekHead & aMetaSeek, const KaxSegment & FirstSegment)
{
	KaxCues * result = NULL;
	KaxSeek * aElt = aMetaSeek.FindFirstOf(KaxCues::ClassInfos);
	
	if (aElt != NULL)
	{
		// get the location and read it
		KaxSeekPosition * aPosition = static_cast<KaxSeekPosition *>(aElt->FindFirstElt(KaxSeekPosition::ClassInfos, false));
		if (aPosition != NULL)
		{
			uint64 CuePosition = FirstSegment.GetGlobalPosition(uint64(*aPosition));
			uint64 CurrentLocation = inDataStream.I_O().getFilePointer();
			int upper = 0;
			EbmlElement *fake;

			inDataStream.I_O().setFilePointer(CuePosition);
			fake = inDataStream.FindNextElement(KaxSegment::ClassInfos.Context, upper, 0xFFFFFFFFL, false);
			if (fake != NULL && EbmlId(*fake) == KaxCues::ClassInfos.GlobalId) {
				result = static_cast<KaxCues *>(fake);
				upper=0;
				result->Read(inDataStream, KaxCues::ClassInfos.Context, upper, fake, false);
			}
			inDataStream.I_O().setFilePointer(CurrentLocation);
		}
	}

	return result;
}

int64 MatroskaReader::FindClusterLocated(int64 aLocation, const KaxSeekHead & aMetaSeek)
{
	int64 result = aLocation;
	KaxSeek *PrevClust = NULL, *NextClust = NULL;
	KaxSeek *aElt = aMetaSeek.FindFirstOf(KaxCluster::ClassInfos);

	while (aElt != NULL)
	{
		if (aElt->Location() < aLocation && (PrevClust == NULL || aElt->Location() > PrevClust->Location()))
		{
			PrevClust = aElt;
		}
		if (aElt->Location() > aLocation && (NextClust == NULL || aElt->Location() < NextClust->Location()))
		{
			NextClust = aElt;
		}
		aElt = aMetaSeek.FindNextOf(*aElt);
	}

	return (PrevClust == NULL)? result: PrevClust->Location();
}
