/****************************************************************************
** libmatroska : parse Matroska files, see http://www.matroska.org/
**
** <file/class description>
**
** Copyright (C) 2002-2003 Steve Lhomme.  All rights reserved.
**
** This file is part of libmatroska.
**
** This file may be distributed under the terms of the Q Public License
** as defined by Trolltech AS of Norway and appearing in the file
** LICENSE.QPL included in the packaging of this file.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding an other license may use this file in accordance with 
** the Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.matroska.org/license/qpl/ for QPL licensing information.
** See http://www.matroska.org/license/gpl/ for GPL licensing information.
**
** Contact license@matroska.org if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

/*!
    \file
    \version \$Id: test6.cpp,v 1.19 2003/04/28 21:55:54 robux4 Exp $
    \brief Test muxing two tracks into valid clusters/blocks/frames
    \author Steve Lhomme     <robux4 @ users.sf.net>
*/

#ifdef _MSC_VER
#include <windows.h>      // for min/max
#endif // _MSC_VER

#include <iostream>

#include "StdIOCallback.h"

#include "EbmlHead.h"
#include "EbmlSubHead.h"
#include "EbmlVoid.h"
#include "FileKax.h"
#include "KaxSegment.h"
#include "KaxTracks.h"
#include "KaxTrackEntryData.h"
#include "KaxTrackAudio.h"
#include "KaxTrackVideo.h"
#include "KaxCluster.h"
#include "KaxClusterData.h"
#include "KaxBlock.h"
#include "KaxBlockData.h"
#include "KaxSeekHead.h"
#include "KaxCues.h"
#include "KaxInfo.h"
#include "KaxInfoData.h"

using namespace LIBMATROSKA_NAMESPACE;
using namespace std;

unsigned int BIN_FILE_SIZE = 15000;
unsigned int TXT_FILE_SIZE = 3000;
const unsigned int BIN_FRAME_SIZE = 1500;
const unsigned int TXT_FRAME_SIZE = 200;
const uint64  TIMECODE_SCALE = 1000000; 

const bool bWriteDefaultValues = false;

/*!
    The first file is a "binary" file with data scaling from 0x00 to 0xFF repeatedly
    The second file is a "text" file with data scaling from 'z' to 'a'
*/
int main(int argc, char **argv)
{
	char *file1;
	char *file2;
	if (argc == 1)
	{
		file1 = "binr.bin";
		// create the files
		StdIOCallback bin_file(file1, MODE_CREATE);
		char c;
		uint64 i;
		for (i=0; i<BIN_FILE_SIZE; i++)
		{
			c = char(i % 0x100);
			bin_file.write(&c, 1);
		}
		bin_file.close();

		file2 = "text.bin";
		StdIOCallback txt_file(file2, MODE_CREATE);
		for (i=TXT_FILE_SIZE; i>0; i--)
		{
			c = char((i % 26) + 'a');
			txt_file.write(&c, 1);
		}
		txt_file.close();
	}
	else if(argc < 3)
	{
		cout << "Syntax: mkvmux [binfile txtfile]" << endl;
		return 1; // error
	}
	else
	{
		file1 = argv[1];
		file2 = argv[2];
	}

    cout << "Creating \"muxed.mkv\"" << endl;

    try {
		// write the head of the file (with everything already configured)
		StdIOCallback out_file("muxed.mkv", MODE_CREATE);

		///// Writing EBML test
		EbmlHead FileHead;

		EDocType & MyDocType = GetChild<EDocType>(FileHead);
		*static_cast<EbmlString *>(&MyDocType) = "matroska";

		EDocTypeVersion & MyDocTypeVer = GetChild<EDocTypeVersion>(FileHead);
		*(static_cast<EbmlUInteger *>(&MyDocTypeVer)) = 1;

		EDocTypeReadVersion & MyDocTypeReadVer = GetChild<EDocTypeReadVersion>(FileHead);
		*(static_cast<EbmlUInteger *>(&MyDocTypeReadVer)) = 1;

		FileHead.Render(out_file, bWriteDefaultValues);

		KaxSegment FileSegment;

		KaxSeekHead MetaSeek;

		KaxInfo & MyInfos = GetChild<KaxInfo>(FileSegment);
		KaxTimecodeScale & TimeScale = GetChild<KaxTimecodeScale>(MyInfos);
		*(static_cast<EbmlUInteger *>(&TimeScale)) = TIMECODE_SCALE;

		KaxDuration & SegDuration = GetChild<KaxDuration>(MyInfos);
		*(static_cast<EbmlFloat *>(&SegDuration)) = 0.0;

		// size is unknown and will always be, we can render it right away
		uint64 SegmentSize = FileSegment.Render(out_file, bWriteDefaultValues);
		
		KaxTracks & MyTracks = GetChild<KaxTracks>(FileSegment);

		
		EbmlVoid Dummy;
		Dummy.SetSize(300); // 300 octets
		Dummy.Render(out_file, bWriteDefaultValues);

		// fill track 1 params
		KaxTrackEntry & MyTrack1 = GetChild<KaxTrackEntry>(MyTracks);

		KaxTrackNumber & MyTrack1Number = GetChild<KaxTrackNumber>(MyTrack1);
		*(static_cast<EbmlUInteger *>(&MyTrack1Number)) = 1;

		KaxTrackUID & MyTrack1UID = GetChild<KaxTrackUID>(MyTrack1);
		*(static_cast<EbmlUInteger *>(&MyTrack1UID)) = 7;

		*(static_cast<EbmlUInteger *>(&GetChild<KaxTrackType>(MyTrack1))) = track_audio;

		KaxCodecID & MyTrack1CodecID = GetChild<KaxCodecID>(MyTrack1);
		MyTrack1CodecID.CopyBuffer((binary*)"Dummy Audio Codec",countof("Dummy Audio Codec"));

		MyTrack1.EnableLacing(true);

		// audio specific params
		KaxTrackAudio & MyTrack1Audio = GetChild<KaxTrackAudio>(MyTrack1);
		
		KaxAudioSamplingFreq & MyTrack1Freq = GetChild<KaxAudioSamplingFreq>(MyTrack1Audio);
		*(static_cast<EbmlFloat *>(&MyTrack1Freq)) = 44100.0;
		MyTrack1Freq.ValidateSize();

		KaxAudioChannels & MyTrack1Channels = GetChild<KaxAudioChannels>(MyTrack1Audio);
		*(static_cast<EbmlUInteger *>(&MyTrack1Channels)) = 2;

		// fill track 2 params
		KaxTrackEntry & MyTrack2 = GetNextChild<KaxTrackEntry>(MyTracks, MyTrack1);

		KaxTrackNumber & MyTrack2Number = GetChild<KaxTrackNumber>(MyTrack2);
		*(static_cast<EbmlUInteger *>(&MyTrack2Number)) = 200;

		KaxTrackUID & MyTrack2UID = GetChild<KaxTrackUID>(MyTrack2);
		*(static_cast<EbmlUInteger *>(&MyTrack2UID)) = 13;

		*(static_cast<EbmlUInteger *>(&GetChild<KaxTrackType>(MyTrack2))) = track_video;

		KaxCodecID & MyTrack2CodecID = GetChild<KaxCodecID>(MyTrack2);
		MyTrack2CodecID.CopyBuffer((binary*)"Dummy Video Codec",countof("Dummy Video Codec"));

		MyTrack2.EnableLacing(false);

		// video specific params
		KaxTrackVideo & MyTrack2Video = GetChild<KaxTrackVideo>(MyTrack2);

		KaxVideoPixelHeight & MyTrack2PHeight = GetChild<KaxVideoPixelHeight>(MyTrack2Video);
		*(static_cast<EbmlUInteger *>(&MyTrack2PHeight)) = 200;

		KaxVideoPixelWidth & MyTrack2PWidth = GetChild<KaxVideoPixelWidth>(MyTrack2Video);
		*(static_cast<EbmlUInteger *>(&MyTrack2PWidth)) = 320;

		uint64 TrackSize = MyTracks.Render(out_file, bWriteDefaultValues);

		MetaSeek.IndexThis(MyTracks, FileSegment);

#ifdef OLD
		// create the Matroska file
		FileMatroska MuxedFile(out_file);

		// add the binary track (audio)
		TrackAudio AudioTrack;
		Track * binID = MuxedFile.CreateTrack( AudioTrack );
		MuxedFile.track_SetName(binID, "Audio test");
		MuxedFile.track_SetLaced(binID);

		TrackInfoAudio aAudioTrack;
		memset(&aAudioTrack, 0, sizeof(aAudioTrack));
		aAudioTrack.SampleRate = 44100;
		aAudioTrack.BitDepth = 8;
		aAudioTrack.ChannelPerSubtrack = 1;
		aAudioTrack.DTS = false;
		aAudioTrack.SecondaryTrack = false;
		aAudioTrack.SubtrackNumber = 1;
		aAudioTrack.ChannelPosition = SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT;

		MuxedFile.Track_SetInfo_Audio(binID, aAudioTrack);

		// add the text track (subtitle)
		TrackSubtitle SubtitleTrack;
		Track * txtID = MuxedFile.CreateTrack( SubtitleTrack );
		MuxedFile.track_SetName(txtID, "Subtitle test");

		// try a video track
		TrackVideo VideoTrack;
		Track * videoID = MuxedFile.CreateTrack( VideoTrack );
		MuxedFile.track_SetName(videoID, "Video test");

		TrackInfoVideo aVideoTrack;
		memset(&aVideoTrack, 0, sizeof(aVideoTrack));

		aVideoTrack.AspectRatio = 1333; // 4:3
		aVideoTrack.Eyes = eye_both;
		aVideoTrack.FourCC[0] = 'X';
		aVideoTrack.FourCC[1] = 'v';
		aVideoTrack.FourCC[2] = 'i';
		aVideoTrack.FourCC[3] = 'D';
		aVideoTrack.Interlaced = false;
		aVideoTrack.xSize = 640;
		aVideoTrack.ySize = 480;

		MuxedFile.Track_SetInfo_Video(videoID, aVideoTrack);

		// first, write the header in the file
		MuxedFile.RenderHead("libmatroska test6.cpp");

		// to verify the muxing is working
		MuxedFile.SetMaxClusterSize(2000);
#endif // OLD

		// creation of the original binary/raw files
		StdIOCallback bin_file(file1, MODE_READ);
		StdIOCallback txt_file(file2, MODE_READ);
		
		bin_file.setFilePointer(0L, seek_end);
		txt_file.setFilePointer(0L, seek_end);
		
		BIN_FILE_SIZE = (unsigned int) bin_file.getFilePointer();
		TXT_FILE_SIZE = (unsigned int) txt_file.getFilePointer();
		
		binary *buf_bin = new binary[BIN_FILE_SIZE];
		binary *buf_txt = new binary[TXT_FILE_SIZE];

		// start muxing (2 binary frames for each text frame)
		bin_file.setFilePointer(0L, seek_beginning);
		txt_file.setFilePointer(0L, seek_beginning);

		bin_file.read(buf_bin, BIN_FILE_SIZE);
		txt_file.read(buf_txt, TXT_FILE_SIZE);

		binary *curs_bin = buf_bin;
		binary *curs_txt = buf_txt;
		unsigned int remain_bin = BIN_FILE_SIZE;
		unsigned int remain_txt = TXT_FILE_SIZE;

	// "manual" filling of a cluster"
	/// \todo whenever a BlockGroup is created, we should memorize it's position
	KaxCues AllCues;

	KaxCluster Clust1;
	Clust1.SetParent(FileSegment); // mandatory to store references in this Cluster
	Clust1.SetPreviousTimecode(0); // the first timecode here
/* no more manual filling possible !!!
	KaxClusterTimecode & MyClusterTimecode = GetChild<KaxClusterTimecode>(Clust1);
	*(static_cast<EbmlUInteger *>(&MyClusterTimecode)) = 0; // base timecode \todo should be automatically computed

	KaxBlockGroup & MyBlockG1 = Clust1.GetNewBlock();
	KaxBlock & MyBlock1 = GetChild<KaxBlock>(MyBlockG1);
	MyBlock1.SetParent(Clust1);

	DataBuffer data1 = DataBuffer((binary *)"test", countof("test"));
	MyBlock1.AddFrame(MyTrack1, 10 * TIMECODE_SCALE, data1);
	AllCues.AddBlockGroup(MyBlockG1);

	KaxBlockGroup & MyBlockG2 = Clust1.GetNewBlock();
	KaxBlock & MyBlock2 = GetChild<KaxBlock>(MyBlockG2);
	MyBlock2.SetParent(Clust1);
	AllCues.AddBlockGroup(MyBlockG2);

	DataBuffer data2 = DataBuffer((binary *)"toast", countof("toast"));
	MyBlock2.AddFrame(MyTrack2, 0 * TIMECODE_SCALE, data2);

	KaxBlockGroup & MyBlockG3 = Clust1.GetNewBlock();
	KaxBlock & MyBlock3 = GetChild<KaxBlock>(MyBlockG3);
	MyBlock3.SetParent(Clust1);

	MyBlock3.AddFrame(MyTrack1, 100 * TIMECODE_SCALE, data1);

	// put some additional data
	KaxReferenceBlock & MyBlockAddPRef3 = GetChild<KaxReferenceBlock>(MyBlockG3);
	MyBlockAddPRef3.SetParentBlock(MyBlockG3);
	MyBlockAddPRef3.SetReferencedBlock(MyBlockG2);
	AllCues.AddBlockGroup(MyBlockG3);

	// automatic filling of a BlockGroup
	// simple frame
	KaxBlockGroup & MyBlockG4 = Clust1.GetNewBlock();
	DataBuffer data4 = DataBuffer((binary *)"titi", countof("titi"));
	MyBlockG4.AddFrame(MyTrack1, 120 * TIMECODE_SCALE, data4);
	MyBlockG4.AddFrame(MyTrack1, 140 * TIMECODE_SCALE, data1); // to test lacing

	// frame with a past reference
	KaxBlockGroup & MyBlockG5 = Clust1.GetNewBlock();
	DataBuffer data5 = DataBuffer((binary *)"tatata", countof("tatata"));
	MyBlockG5.AddFrame(MyTrack1, 200 * TIMECODE_SCALE, data5, MyBlockG4);

	// frame with a past & future reference
	KaxBlockGroup & MyBlockG6 = Clust1.GetNewBlock();
	DataBuffer data6 = DataBuffer((binary *)"tetetete", countof("tetetete"));
	MyBlockG6.AddFrame(MyTrack1, 180 * TIMECODE_SCALE, data6, MyBlockG1, MyBlockG5);
	AllCues.AddBlockGroup(MyBlockG6);
*/
	// automatic filling of a Cluster
	// simple frame
	KaxBlockGroup *MyNewBlock, *MyLastBlockTrk1 = NULL, *MyLastBlockTrk2 = NULL, *MyNewBlock2;
	DataBuffer data7 = DataBuffer((binary *)"tototototo", countof("tototototo"));
	Clust1.AddFrame(MyTrack1, 250 * TIMECODE_SCALE, data7, MyNewBlock);
	if (MyNewBlock != NULL)
		MyLastBlockTrk1 = MyNewBlock;
	Clust1.AddFrame(MyTrack1, 270 * TIMECODE_SCALE, data7, MyNewBlock); // to test lacing
	if (MyNewBlock != NULL) {
		MyLastBlockTrk1 = MyNewBlock;
	} else {
		MyLastBlockTrk1->SetBlockDuration(50 * TIMECODE_SCALE);
	}
	Clust1.AddFrame(MyTrack2, 23 * TIMECODE_SCALE, data7, MyNewBlock); // to test with another track

	// add the "real" block to the cue entries
	AllCues.AddBlockGroup(*MyLastBlockTrk1);

	// frame for Track 2
	DataBuffer data8 = DataBuffer((binary *)"tttyyy", countof("tttyyy"));
	Clust1.AddFrame(MyTrack2, 107 * TIMECODE_SCALE, data8, MyNewBlock, *MyLastBlockTrk2);

	AllCues.AddBlockGroup(*MyNewBlock);

	// frame with a past reference
	Clust1.AddFrame(MyTrack1, 300 * TIMECODE_SCALE, data8, MyNewBlock, *MyLastBlockTrk1);

	// frame with a past & future reference
	if (MyNewBlock != NULL) {
		if (Clust1.AddFrame(MyTrack1, 280 * TIMECODE_SCALE, data8, MyNewBlock2, *MyLastBlockTrk1, *MyNewBlock)) {
			MyNewBlock2->SetBlockDuration(20 * TIMECODE_SCALE);
			MyLastBlockTrk1 = MyNewBlock2;
		} else {
			printf("Error adding a frame !!!");
		}
	}

	AllCues.AddBlockGroup(*MyLastBlockTrk1);
#ifdef OLD
		while (remain_bin !=0 || remain_txt != 0)
		{
			// a text chunck
			if (remain_txt !=0) {
				MuxedFile.AddFrame(txtID, 0, curs_txt, min(TXT_FRAME_SIZE, remain_txt));
				curs_txt += min(TXT_FRAME_SIZE,remain_txt);
				remain_txt -= min(TXT_FRAME_SIZE,remain_txt);
			}
			// a binary chunck
			if (remain_bin !=0) {
//				MuxedFile.AddFrame(binID, 0, curs_bin, min(BIN_FRAME_SIZE,remain_bin));
				MuxedFile.AddFrame(videoID, 0, curs_bin, min(BIN_FRAME_SIZE, remain_bin));
				curs_bin += min(BIN_FRAME_SIZE,remain_bin);
				remain_bin -= min(BIN_FRAME_SIZE,remain_bin);
			}
			// a text chunck
			if (remain_txt !=0) {
				MuxedFile.AddFrame(txtID, 0, curs_txt, min(TXT_FRAME_SIZE, remain_txt));
				curs_txt += min(TXT_FRAME_SIZE,remain_txt);
				remain_txt -= min(TXT_FRAME_SIZE,remain_txt);
			}
		}
#endif // OLD

		// simulate the writing of the stream :
		// - write an empty element with enough size for the cue entry
		// - write the cluster(s)
		// - seek back in the file and write the cue entry over the empty element

		uint64 ClusterSize = Clust1.Render(out_file, AllCues, bWriteDefaultValues);
		Clust1.ReleaseFrames();
		MetaSeek.IndexThis(Clust1, FileSegment);

		KaxCluster Clust2;
		Clust2.SetParent(FileSegment); // mandatory to store references in this Cluster
		Clust2.SetPreviousTimecode(300 * TIMECODE_SCALE); // the first timecode here
	
		Clust2.AddFrame(MyTrack1, 350 * TIMECODE_SCALE, data7, MyNewBlock, *MyLastBlockTrk1);
		
		AllCues.AddBlockGroup(*MyNewBlock);

		ClusterSize += Clust2.Render(out_file, AllCues, bWriteDefaultValues);
		Clust2.ReleaseFrames();

// older version, write at the end		AllCues.Render(out_file);
		uint32 CueSize = AllCues.Render(out_file, bWriteDefaultValues);
		MetaSeek.IndexThis(AllCues, FileSegment);

		// \todo put it just before the Cue Entries
		uint32 MetaSeekSize = Dummy.ReplaceWith(MetaSeek, out_file, bWriteDefaultValues);

#ifdef VOID_TEST
		MyInfos.VoidMe(out_file);
#endif // VOID_TEST

		// let's assume we know the size of the Segment element
		// the size of the FileSegment is also computed because mandatory elements we don't write ourself exist
		if (FileSegment.ForceSize(SegmentSize - FileSegment.HeadSize() + MetaSeekSize + TrackSize + ClusterSize + CueSize)) {
			FileSegment.OverwriteHead(out_file);
		}

		delete[] buf_bin;
		delete[] buf_txt;

#ifdef OLD
		MuxedFile.Close(1000); // 1000 ms
#endif // OLD
		out_file.close();
    }
    catch (exception & Ex)
    {
		cout << Ex.what() << endl;
    }

    return 0;
}
